package net.pricefx.service.impl;

import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;
import com.googlecode.genericdao.search.Filter;
import net.pricefx.common.api.PriceFxInterface;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.function.Function;

public class AdvancedFilterUtil {
    @SuppressWarnings({ "rawtypes", "unchecked" })
	private static Filter buildFilter(Map<String,?> criteria, Map<String, String> fieldTranslation, Function<String, Integer> metaCallback) {
		String opCode = (String) criteria.get(PriceFxInterface.PARAM_FILTER_CRITERIA_OPERATOR);
		int opId = matchOperator(opCode);
		if(opId >= 0) {
			Filter f = new Filter();
			f.setOperator(opId);
			String fieldName = (String) criteria.get("fieldName");
			String valuePrefix = "";
			String valueSuffix = "";
			if(fieldTranslation != null && fieldTranslation.containsKey(fieldName)) {
				fieldName = fieldTranslation.get(fieldName);
			}
			if(fieldName == null || fieldName.startsWith("_") || fieldName.startsWith("$"))
				return null;

			f.setProperty(fieldName);

			Object criteriaValue = criteria.get("value");

			// sgwt sends [fields equals null)... need to replace with [field isNull]
			if (criteriaValue == null && opId == Filter.OP_EQUAL) {
			    f.setOperator(Filter.OP_NULL);
			    f.setValue(null);
			} else if (criteriaValue == null && opId == Filter.OP_NOT_EQUAL) {
                f.setOperator(Filter.OP_NOT_NULL);
                f.setValue(null);
            }
			
			if("isNull".equals(opCode)) {
				f.setOperator(Filter.OP_NULL);
			    f.setValue(null);
			}
			
			if("notNull".equals(opCode)) {
				f.setOperator(Filter.OP_NOT_NULL);
			    f.setValue(null);
			}			
			
			if("isBlank".equals(opCode)) {
				f = Filter.or(Filter.equal(fieldName, ""),Filter.isNull(fieldName));
			}		
			
			if("notBlank".equals(opCode)) {
				f = Filter.not(Filter.or(Filter.equal(fieldName, ""),Filter.isNull(fieldName)));
			}
			
			if ("greaterThan".equals(opCode) || "lessThan".equals(opCode) || "greaterOrEqual".equals(opCode) || "lessOrEqual".equals(opCode)){
				if (criteriaValue instanceof String) {
					f.setValue(criteriaValue.toString().trim());
				} else {
					f.setValue(criteriaValue);
				}
			} else if("startsWith".equals(opCode) || "iStartsWith".equals(opCode) || "iNotStartsWith".equals(opCode) || "notStartsWith".equals(opCode)){
				f.setValue(String.format("%s%s%s%%", valuePrefix, criteriaValue, valueSuffix));
			} else if("endsWith".equals(opCode) || "iEndsWith".equals(opCode)|| "iNotEndsWith".equals(opCode) || "notEndsWith".equals(opCode)){
				f.setValue(String.format("%%%s%s%s", valuePrefix, criteriaValue, valueSuffix));
			} else if("contains".equals(opCode) || "iContains".equals(opCode) || "iNotContains".equals(opCode) || "notContains".equals(opCode)){
				// quite a few sgwt operators are mapped to LIKE/ILIKE - see matchOperator()...
				// for this reason we want to make sure we don't expand the filter value further as '%s<value>%s' (as filters are converted to maps and back).
				// obviously this rules out criteria like (prop like 'abd%') and (prop like '%abc') but that's fine as the wildcards there don't add anything.
				if (criteriaValue instanceof String 
						&& (criteriaValue.toString().startsWith("%") || criteriaValue.toString().endsWith("%"))){
						
					f.setValue(criteriaValue);
				} else {
					f.setValue(String.format("%%%s%s%s%%", valuePrefix, criteriaValue, valueSuffix));
				}
			} else if("containsPattern".equals(opCode) || "iContainsPattern".equals(opCode) || "iNotContainsPattern".equals(opCode) || "notContainsPattern".equals(opCode)){
				f.setValue(String.format("%%%s%s%s%%", valuePrefix, criteriaValue, valueSuffix));
			} else if("equals".equals(opCode) || "iEquals".equals(opCode)){
				if (criteriaValue instanceof Number && valuePrefix.isEmpty() && valueSuffix.isEmpty()) {
					f.setValue(criteriaValue);
				} else {
					f.setValue(String.format("%s%s%s", valuePrefix, criteriaValue, valueSuffix));
				}
			} else {
				f.setValue(criteriaValue);
			}
			
			if ("iNotContains".equals(opCode) ||
				"iNotContainsPattern".equals(opCode) ||
				"iNotStartsWith".equals(opCode) ||
				"iNotEndsWith".equals(opCode) ||
				"iNotEqual".equals(opCode) ||
				"notContains".equals(opCode) ||
				"notContainspattern".equals(opCode) ||
				"notStartsWith".equals(opCode) ||
				"notEndsWith".equals(opCode)) {
				f = Filter.not(f);
				
			}
			
			if(criteriaValue != null) {
				if("iNotEqual".equals(opCode) ||
					"notEqual".equals(opCode) ||
					"iNotContains".equals(opCode) || 
					"notContains".equals(opCode) ||
					"notContainspattern".equals(opCode)) {
					f = Filter.or(f,Filter.isNull(fieldName));
				}
			}
		
			if ("inSet".equals(opCode) ||
				"notInSet".equals(opCode)) {
				// check if the filter's value happens to contain null value(s) and if yes, filter it/them out and explicitly add extra OP_NULL/OP_NOT_NULL filter
				// as SQL IN operator couldn't work out as expected otherwise
				
				if(criteriaValue != null && criteriaValue.getClass().isArray()) {
					criteriaValue = new ArrayList(Arrays.asList(criteriaValue));
				} else if (criteriaValue instanceof Collection) {
					int origSize = ((Collection<?>) criteriaValue).size();
					criteriaValue = Collections2.filter((Collection<?>) criteriaValue, new Predicate<Object>() {
						@Override
						public boolean apply(Object item) {
							return item != null; // filter out nulls
						}
					});
					if (origSize != ((Collection<?>) criteriaValue).size()) { // criteriaValue contained null(s)
						f.setValue(criteriaValue);
						f = "inSet".equals(opCode) ? Filter.or(f, Filter.isNull(f.getProperty())) : Filter.and(f, Filter.isNotNull(f.getProperty()));
					}
				} else if (criteriaValue != null) {
					f.setValue(new ArrayList<>(Collections.singletonList(criteriaValue)));
				}
			}
				
			return f;
		} else if("iBetweenInclusive".equals(opCode) || "betweenInclusive".equals(opCode)) {
			Filter start = new Filter();
			start.setProperty((String) criteria.get("fieldName"));
			start.setOperator(Filter.OP_GREATER_OR_EQUAL);
			start.setValue(criteria.get("start"));
			
			Filter end = new Filter();
			end.setProperty((String) criteria.get("fieldName"));
			end.setOperator(Filter.OP_LESS_OR_EQUAL);
			end.setValue(criteria.get("end"));
			
			return Filter.and(start,end);
		} else if("iBetween".equals(opCode) || "between".equals(opCode)) {
			Filter start = new Filter();
			start.setProperty((String) criteria.get("fieldName"));
			start.setOperator(Filter.OP_GREATER_THAN);
			start.setValue(criteria.get("start"));
			
			Filter end = new Filter();
			end.setProperty((String) criteria.get("fieldName"));
			end.setOperator(Filter.OP_LESS_THAN);
			end.setValue(criteria.get("end"));
			
			return Filter.and(start,end);
						
		} else {
			return null;
		}
	}

	public static Filter buildFilterFromAdvancedCriterion(Map<String,?> advCriterion, Map<String, String> fieldTranslation) {
		return buildFilterFromAdvancedCriterion(advCriterion, fieldTranslation, null);
	}
		@SuppressWarnings({ "unchecked" })
	public static Filter buildFilterFromAdvancedCriterion(Map<String,?> advCriterion, Map<String, String> fieldTranslation, Function<String, Integer> metaCallback) {
		if (advCriterion == null) {
			throw new IllegalArgumentException("criterion argument cannot be null");
		}
		
		List<Map<String,Object>> criteria = (List<Map<String,Object>>) advCriterion.get(PriceFxInterface.PARAM_FILTER_CRITERIA);
		
		if (criteria != null){
			String operator = (String)(advCriterion.get(PriceFxInterface.PARAM_FILTER_CRITERIA_OPERATOR));
			return buildFilterFromAdvancedCriteria(criteria, operator, fieldTranslation, metaCallback);
		} else{
			// simple advanced filter
			return buildFilter(advCriterion, fieldTranslation, metaCallback);
		}
	}

	private static Filter buildFilterFromAdvancedCriteria(List<Map<String,Object>> criteria, String operator, Map<String, String> fieldTranslation) {
		return buildFilterFromAdvancedCriteria(criteria, operator, fieldTranslation, null);
	}

	private static Filter buildFilterFromAdvancedCriteria(List<Map<String,Object>> criteria, String operator, Map<String, String> fieldTranslation, Function<String, Integer> metaCallback) {
		if (criteria == null) {
			throw new IllegalArgumentException("criteria and operator cannot be null");
		}
		// operator is sometimes null in SGWT generated criteria; assume 'and' if so:
		if (operator == null){
			operator = "and";
		}
		List<Filter> filters = new ArrayList<>();
		for (Map<String,Object> criterion : criteria) {
			// TODO: make sure, i.e. test that changing this condition does not screw up the DS protocol when processing client filter requests
			if (criterion.containsKey(PriceFxInterface.PARAM_FILTER_CRITERIA)) {
				filters.add(buildFilterFromAdvancedCriterion(criterion, fieldTranslation, metaCallback));
			}
			else {
				Filter f = buildFilter(criterion, fieldTranslation, metaCallback);
				if (f != null) {
					filters.add(f);
				}
			}
		}
		if (filters.isEmpty()) 
			return null;
		
		//purge empty/null filters from the list
		if(filters.contains(null)) {
			for (Iterator<Filter> iterator = filters.iterator(); iterator.hasNext(); ) {
			    Filter f = iterator.next();
			    if(f==null) {
			        iterator.remove();
			    }
			}			
		}
		
		if("and".equals(operator) || "or".equals(operator)) {
			if (filters.size() > 1){	// we are getting AND of 1 filter from smartgwt - ignore the superfluous AND/OR in this case
				Filter result = ("and".equals(operator))? Filter.and() : Filter.or();
				result.setValue(filters);
				return result;
			}
			else return filters.get(0);
			
		} else if ("not".equals(operator)) { // the 'not' operator means: all subcriteria (criterion.criteria) are false
			if(filters.size() == 1) {
				return Filter.not(filters.get(0));
			} else {
				Filter result = Filter.and();
				for (int i = 0; i < filters.size(); i++) {
					filters.set(i, Filter.not(filters.get(i)));
				}
				result.setValue(filters);
				return result;
			}			
		} else {
			throw new UnsupportedOperationException("AdvancedFilterUtil.buildFilterFromAdvancedCriteria - operator isn't supported yet: " + operator);
		}
	}

	private static int matchOperator(String opCode) {
		//we don't care about case insensitive
//		if(opCode.startsWith("i"))
//			opCode = opCode.substring(1,2).toLowerCase() + opCode.substring(2);
//		
		if("equals".equals(opCode))
			return Filter.OP_EQUAL;
		else if ("notEqual".equals(opCode))
			return Filter.OP_NOT_EQUAL;
		else if ("greaterThan".equals(opCode))
			return Filter.OP_GREATER_THAN;
		else if ("lessThan".equals(opCode))
			return Filter.OP_LESS_THAN;
		else if ("greaterOrEqual".equals(opCode))
			return Filter.OP_GREATER_OR_EQUAL;
		else if ("lessOrEqual".equals(opCode))
			return Filter.OP_LESS_OR_EQUAL;		
		else if ("contains".equals(opCode))
			return Filter.OP_LIKE;		
		else if ("containspattern".equals(opCode))
			return Filter.OP_LIKE;		
		else if ("startsWith".equals(opCode))
			return Filter.OP_LIKE;		
		else if ("endsWith".equals(opCode))
			return Filter.OP_LIKE;
		else if ("isNull".equals(opCode))
			return Filter.OP_NULL;		
		else if ("notNull".equals(opCode))
			return Filter.OP_NOT_NULL;
		else if ("inSet".equals(opCode))
			return Filter.OP_IN;
		else if ("notInSet".equals(opCode))
			return Filter.OP_NOT_IN;
		else if ("iEquals".equals(opCode))
			return Filter.OP_EQUAL;		
		else if ("iContains".equals(opCode))
			return Filter.OP_ILIKE;		
		else if ("iContainsPattern".equals(opCode))
			return Filter.OP_ILIKE;		
		else if ("iStartsWith".equals(opCode))
			return Filter.OP_ILIKE;		
		else if ("iEndsWith".equals(opCode))
			return Filter.OP_ILIKE;
		else if ("iNotContains".equals(opCode))
			return Filter.OP_ILIKE;		
		else if ("iNotContainsPattern".equals(opCode))
			return Filter.OP_ILIKE;		
		else if ("iNotStartsWith".equals(opCode))
			return Filter.OP_ILIKE;		
		else if ("iNotEndsWith".equals(opCode))
			return Filter.OP_ILIKE;
		else if ("iNotEqual".equals(opCode))
			return Filter.OP_ILIKE;	
		else if ("notContains".equals(opCode))
			return Filter.OP_LIKE;		
		else if ("notContainsPattern".equals(opCode))
			return Filter.OP_LIKE;		
		else if ("notStartsWith".equals(opCode))
			return Filter.OP_LIKE;		
		else if ("notEndsWith".equals(opCode))
			return Filter.OP_LIKE;	
		else 
			return -1; //Not supported operation
	}
	
	public static boolean isAdvancedFilter(Map<String,?> filters) {
		if(filters == null)
			return false;
		
		if (filters.containsKey(PriceFxInterface.PARAM_FILTER_CRITERIA)
				|| "AdvancedCriteria".equals(filters.get(PriceFxInterface.PARAM_FILTER_CRITERIA_CONSTRUCTOR))
				|| (filters.containsKey(PriceFxInterface.PARAM_FILTER_CRITERIA_FIELDNAME) &&
						filters .containsKey(PriceFxInterface.PARAM_FILTER_CRITERIA_OPERATOR))) {
			return true;
		} else {
			return false;
		}
	}

	public static Map<String,Object> toAdvancedCriterion(int op, Filter... filter) {
		if(op != Filter.OP_AND && op != Filter.OP_OR && op != Filter.OP_NOT) {
			throw new UnsupportedOperationException("Cannot create adv. criteria from such Filter object yet:" + filter);
		}
		
		Map<String,Object> advFilter = new HashMap<String, Object>();
		if(filter == null || filter.length == 0) return advFilter;
		
		List<Map<String, Object>> criteria = new ArrayList<>();
		
		for(Filter f : filter) {
			if(f.getOperator() == Filter.OP_AND || f.getOperator() == Filter.OP_OR ||
					f.getOperator() == Filter.OP_NOT) {
				List<?> filterValuesAsList = f.getValuesAsList();
				List<Filter> filters = new ArrayList<>();
				for (Object fValue : filterValuesAsList) {
					if(fValue instanceof Filter) {
						filters.add((Filter) fValue);
					}
				}
				criteria.add(toAdvancedCriterion(f.getOperator(), filters.toArray(new Filter[0])));
			} else {
				Map<String,Object> crit = new HashMap<String, Object>();
				crit.put(PriceFxInterface.PARAM_FILTER_CRITERIA_FIELDNAME, f.getProperty());
				crit.put(PriceFxInterface.PARAM_FILTER_CRITERIA_VALUE, f.getValue());
				String opCode = toACOperator(f.getOperator());
				if(opCode != null) {
					crit.put(PriceFxInterface.PARAM_FILTER_CRITERIA_OPERATOR, opCode);
					criteria.add(crit);
				}
			}
		}
		
		if(criteria.size() > 0) {
			advFilter.put(PriceFxInterface.PARAM_FILTER_CRITERIA_CONSTRUCTOR, "AdvancedCriteria");
			advFilter.put(PriceFxInterface.PARAM_FILTER_CRITERIA_OPERATOR, Filter.OP_AND == op ? "and" : Filter.OP_OR == op ? "or" : "not");
			advFilter.put(PriceFxInterface.PARAM_FILTER_CRITERIA, criteria);
		}
		
		return advFilter;
	}
	
	public static Map<String,Object> toAdvancedCriterion(Filter filter) {
		if(filter == null) {
			return new HashMap<String, Object>();
		} else if(filter.getOperator() == Filter.OP_AND || filter.getOperator() == Filter.OP_OR ||
				filter.getOperator() == Filter.OP_NOT) {
			List<?> filterValuesAsList = filter.getValuesAsList();
			List<Filter> filters = new ArrayList<>();
			for (Object fValue : filterValuesAsList) {
				if(fValue instanceof Filter) {
					filters.add((Filter) fValue);
				}
			}
			return toAdvancedCriterion(filter.getOperator(), filters.toArray(new Filter[0]));
		} else {
			return toAdvancedCriterion(Filter.OP_AND, filter);
		}
	}
	
	private static String toACOperator(int operator) {

		if(Filter.OP_EQUAL == operator)
			return "equals";
		else if (Filter.OP_NOT_EQUAL == operator)
			return "notEqual";
		else if (Filter.OP_GREATER_THAN == operator)
			return "greaterThan";
		else if (Filter.OP_LESS_THAN == operator)
			return "lessThan";
		else if (Filter.OP_GREATER_OR_EQUAL == operator)
			return "greaterOrEqual";
		else if (Filter.OP_LESS_OR_EQUAL == operator)
			return "lessOrEqual";		
		else if (Filter.OP_LIKE == operator)
			return "contains";		
		else if (Filter.OP_NULL == operator)
			return "isNull";		
		else if (Filter.OP_NOT_NULL == operator)
			return "notNull";
		else if (Filter.OP_IN == operator)
			return "inSet";
		else if (Filter.OP_NOT_IN == operator)
			return "notInSet";	
		else if (Filter.OP_ILIKE == operator)
			return "iContains";		
	
		else 
			return null; //Not supported operation
	}	
}
